/* eslint-disable max-lines-per-function, max-lines */
import 'isomorphic-unfetch';
import isNode from 'detect-node';

import getLocalEndpoint from './getLocalEndpoint';
import getTimeout from './getTimeout';
import formatArguments from './formatArguments';
import getClient from './getClient';
import errors from './errors';
import { mock, clearMocks, getMockedResponse } from './mock';
import { transformResponse } from './transform';

// Default. Not public. No context required, peer to peer to endpoint - does not support uploads.
const ENDPOINT_LOCAL = getLocalEndpoint();

// Public endpoint. Public. Context required.
const ENDPOINT_PUBLIC = 'https://rpc.bestprice.gr/backend';

// Local, pointing to local rpx-proxy. Context required.
const ENDPOINT_DEV = 'https://local.bestprice.gr:3002/backend';

// If true, append the method to the URI so it's easier to debug.
// NOT when in node context since it will break everything.
const APPEND_METHOD = !isNode;

// e.g React-native env
const HAS_WINDOW = (
  typeof window !== 'undefined'
  && typeof window !== 'undefined'
  && typeof window.navigator.userAgent !== 'undefined'
);

// Default for node is local for browser is public
let ENDPOINT = isNode ? ENDPOINT_LOCAL : ENDPOINT_PUBLIC;
let CONTEXT = isNode ? global.process.env.RPC_CONTEXT : null;
let HOST = null;

const escHandler = event => {
  if (event.keyCode === 27) {
    event.preventDefault();
  }
};

// Apparently Safari does not play well with XHR requests and CORS.
// To that end we shall use this flag to ignore onProgress for safari.
// TODO: figure out how to get around it.
const IS_SAFARI = !isNode && HAS_WINDOW
  && (window.safari
    || (window.navigator.userAgent.indexOf('Safari') !== -1 && window.navigator.userAgent.indexOf('Chrome') === -1));

const defaultOptions = {
  onProgress: null,
  timeout: null,
  transform: null,
  transformResponse: null,
  transformRequest: null,
};

const error = (type, data = {}, method, args) => {
  console.warn(`[fetch-rpc] Error: ${type}`);

  let res = {
    type,
    method,
    args,
  };

  // Include data
  res = Object.assign(res, data);

  // Provide stack on generic errors only
  if (type === errors.GENERIC) {
    res.error = new Error(type);
  }

  res.toString = () => res.faultString || res.error || res.proxyError;

  return res;
};

const fetchRPCWithEndpoint = (endpoint = ENDPOINT, method, ...args) => {
  let dry = null;

  const isUsingLocalEndpoint = endpoint === ENDPOINT_LOCAL;

  if (typeof method === 'boolean') {
    dry = method;
    method = args.shift();
  }

  if (dry === true) {
    console.log(`${method}: ${JSON.stringify(args)})`);
    return Promise.resolve(dry);
  }

  // Append the method in endpoint URI so we can debug easier later.
  if (APPEND_METHOD) {
    endpoint = [endpoint, '/', method.replace(/^bestprice\./, '')].join('');
  }

  if (!method) {
    return error(errors.METHOD_MISSING, {
      args,
    });
  }

  let options = Object.assign({}, defaultOptions);

  // Check if  the last argument is an options object
  let lastArgument = args[args.length - 1];

  if (typeof lastArgument === 'object'
      && typeof defaultOptions[Object.keys(lastArgument)[0]] !== 'undefined') {
    // Remove from args
    lastArgument = args.pop();

    options = Object.assign({},
      defaultOptions,
      lastArgument
    );
  }

  let body;

  // Kep the files ref here
  let filesGlobal;

  // or now off
  // https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS
  // let headers = {
  //   'x-t': Date.now(), // Carry time for extra validation
  // };

  let headers = {};

  // Setting this to browser would cause a warning
  if (isNode) {
    headers.referer = 'https://www.bestprice.gr';
  }

  // For all cases but local where we just proxy to RPC server directly
  if (!isUsingLocalEndpoint) {
    const supportsFormData = !isNode;

    // FormData is not natively supported by node.
    // let's do it the lazy way.
    const data = supportsFormData
      ? new FormData()
      : {};

    const {
      formattedArgs,
      files,
    } = formatArguments(args, options);

    // If we have a mock, return it as a promise
    const mockedResponse = getMockedResponse({
      m: method,
      a: formattedArgs,
    });

    if (typeof mockedResponse !== 'undefined') {
      return Promise.resolve(transformResponse(mockedResponse, options));
    }

    // ref for later
    filesGlobal = files;

    if (supportsFormData) {
      data.append('m', method);
      data.append('a', JSON.stringify(formattedArgs));
      data.append('client', JSON.stringify(getClient()));

      if (CONTEXT) {
        data.append('ctx', CONTEXT);
      }

      // Pass files
      files.forEach((file, index) => data.append(`file${index}`, file));

      body = data;
    } else {
      data.m = method;
      data.a = JSON.stringify(formattedArgs);

      if (CONTEXT) {
        data.ctx = CONTEXT;
      }

      body = JSON.stringify(data);

      headers = {
        'Content-type': 'application/json',
        'Content-Length': body.byteLength, // This is neded since we pass a buffer not a string
      };
    }
  } else {
    const {
      formattedArgs,
    } = formatArguments(args, options);

    const data = {
      m: method,
      a: formattedArgs,
      ctx: CONTEXT,
    };

    // If we have a mock, return it as a promise
    const mockedResponse = getMockedResponse({
      m: method,
      a: formattedArgs,
    });

    if (typeof mockedResponse !== 'undefined') {
      return Promise.resolve(transformResponse(mockedResponse, options));
    }

    // In case we need to point to custom host
    if (HOST) {
      data.host = HOST;
    }

    body = JSON.stringify(data);

    // If node, convert to iso and send buffer
    if (isNode) {
      const toISO = require('./toISO');
      body = toISO(body);

      headers = {
        'Content-type': 'application/json',
        'Content-Length': body.byteLength, // This is neded since we pass a buffer not a string
      };
    }
  }

  if (!isNode && HAS_WINDOW) {
    window.addEventListener('keydown', escHandler);
  }

  // If timeout is not provided as an option from the user, then use the one provided by getTimeout ()
  if (!options.timeout) {
    options = Object.assign(options, {
      timeout: getTimeout({ files: filesGlobal }),
    });
  }

  // ref to timer;
  let timer;

  try {
    if (options.onProgress && !isNode && !IS_SAFARI) {
      const xhr = new XMLHttpRequest();

      xhr.upload.addEventListener('progress', event => {
        if (event.lengthComputable) {
          const { total, loaded } = event;
          const pct = Math.round(loaded / total * 100);

          options.onProgress({ pct, loaded, total });
        }
      });

      xhr.withCredentials = true;

      xhr.open('post', endpoint);

      xhr.send(body);

      const promise = new Promise((resolve, reject) => {
        xhr.addEventListener('error', () => reject(error(errors.GENERIC, error, method, args)));

        xhr.addEventListener('load', () => {
          clearTimeout(timer);

          try {
            if (HAS_WINDOW) {
              window.removeEventListener('keydown', escHandler);
            }

            const json = JSON.parse(xhr.responseText);

            if (json && json.faultString) {
              return reject(error(errors.RPC, json, method, args));
            }

            if (json && json.proxyError) {
              return reject(error(errors.PROXY, json.proxyError, method, args));
            }

            return resolve(transformResponse(json, options));
          } catch (genericError) {
            if (HAS_WINDOW) {
              window.removeEventListener('keydown', escHandler);
            }

            return reject(error(errors.GENERIC, genericError, method, args));
          }
        });
      });

      // Enhance with abort method
      promise.abort = () => {
        clearTimeout(timer);
        xhr.abort();
      };

      // Timer
      timer = setTimeout(() => promise.abort(), options.timeout);

      return promise;
    }

    // Default method
    const abortController = !isNode && typeof AbortController !== 'undefined'
      // eslint-disable-next-line
      ? new AbortController()
      : null;

    // Node needs this so that aux calls can be routed properly
    if (isNode) {
      // eslint-disable-next-line
      endpoint += '?method=' + method;
    }

    const promise = new Promise((resolve, reject) => {
      const fetchRromise = fetch(endpoint, {
        method: 'post',
        credentials: 'include',
        mode: 'cors', // Tho, this is the default already
        cache: 'no-store', // https://developer.mozilla.org/en-US/docs/Web/API/Request/cache
        body,
        headers,
        signal: abortController ? abortController.signal : undefined,
      });

      fetchRromise.then(res => {
        clearTimeout(timer);

        if (!isNode && HAS_WINDOW) {
          window.removeEventListener('keydown', escHandler);
        }

        const responseOk = (res.ok || res.status === 200 || res.status === 400) && res.status !== 408;

        // This is not a valid assumption, but it's needed to pass the error properly back
        const isErrorResponseFromGo = res.status === 500;

        if (!responseOk && !isErrorResponseFromGo) {
          return reject(error(errors.NETWORK, res, method, args));
        }

        return res.json();
      })
        .then(data => {
          if (data && data.faultString) {
            return reject(error(errors.RPC, data, method, args));
          }

          if (data && data.proxyError) {
            return reject(error(errors.PROXY, data, method, args));
          }

          return resolve(transformResponse(data, options));
        });
    });

    // Timer
    timer = setTimeout(() => {
      promise.abort();
    }, options.timeout);

    // The abort method will have to be carried to the promise and the inner promises
    const abort = () => {
      clearTimeout(timer);

      if (abortController) {
        abortController.abort();
      } else {
        console.warn('Abort is N/A');
      }
    };

    promise.abort = abort;

    return promise;
  } catch (genericError) {
    // Global error handler
    if (!isNode && HAS_WINDOW) {
      window.removeEventListener('keydown', escHandler);
    }

    if (genericError.name === 'AbortError') {
      return Promise.reject(error(errors.ABORTED, {}, method, args));
    }

    return Promise.reject(error(errors.GENERIC, genericError, method, args));
  }
};

const local = () => {
  setEndpoint(ENDPOINT_LOCAL);
  return fetchRPC;
};

const dev = context => {
  setEndpoint(ENDPOINT_DEV);
  setContext(context);
  return fetchRPC;
};

const pub = context => {
  setEndpoint(ENDPOINT_PUBLIC);
  setContext(context);

  return fetchRPC;
};

const setEndpoint = newEndpoint => {
  ENDPOINT = newEndpoint;
  return fetchRPC;
};

const setHost = host => {
  HOST = host;
  return fetchRPC;
};

const setContext = newContext => {
  CONTEXT = newContext;
  return fetchRPC;
};

// Safari has issues here: arguments: TypeError: 'arguments', 'callee', and 'caller' cannot be accessed in this context.
const wait = ms => (...props) =>
  Promise.all([
    Reflect.apply(fetchRPC, undefined, Array.isArray(props) ? props : []),
    new Promise(resolve => setTimeout(resolve, ms)),
  ]);

const fetchRPC = (...input) => fetchRPCWithEndpoint(ENDPOINT, ...input);

Object.assign(fetchRPC, {
  local,
  pub,

  dev,
  setContext,
  setEndpoint,
  setHost,
  errors,
  formatArguments,
  wait,

  mock,
  clearMocks,

  // It might better to spin this off to a different file later since
  // this results in extra 10k in size.
});

// This has to be exported like this, and not as /hook so that mocks can be shared
export { default as useRPC } from './hook';

export default fetchRPC;
